/*
vdr2divx - full automated conversion of vdr recordings into (divx)avi files
Copyright (C) 2003 Martin Hoffmann / ma.hoff@t-online.de / ma.hoff@bigfoot.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 

For the complete actual license take a look at www.gnu.org/copyleft/gpl.html

*/

#include <String.h>
#include <iostream>
#include <GetOpt.h>

#include "FileExist.h"
#include "Video.h"
#include "Config.h"
#include "System.h"
#include "modules/dmx_dsjar.h"
#include "modules/aen_lame.h"
#include "modules/ven_mencoder.h"

//String IntToString (unsigned int Zahl, int Stellen);
bool CheckEnv ();
int CalcVBitrate(int CDSize, int NumCDs, int Files, int frames, int ABitrate);
void Usage(void);
void UsageDebug(void);


int main (int argc, char *argv[])
{
    if (argc<=1)	//quick check for too few arguments!
    {		// one argument is at least required !!!
	Usage();
	return 1;
    }
    
    if (!CheckEnv())
    {
      cerr << "*** Fatal: Exiting due to major configuration error - see above for details.\n";
      return 0;
    }
    
    cVideo SrcVideo;	// SrcVideo Class - contains all params about source!
    cVideo DstVideo;	// all params about Destination File(s)

    cConfig Config;	// Overall 2divx Configuration
    if (!Config.ConfFile())	// Read in everything from 2divx.conf
    {
	cerr << "*** Fatal: Exiting due to errors in 2divx.conf - see above.\n";
	return 0;
    }
//    Config.Dump();
// various temporary variables 
    String FileName;
    String Number;
    String Str;
    String Str2;
    String Str3;
    String Name;
    String Ext;
    int i;
    int Hours;
    int Minutes;
    int Seconds;
    int SingleFrames;
    int VBitrate;
    ifstream InFile;
    char OneChar;


    GetOpt parser(argc, argv, "dD:"); /* initialize the parser */
//    int c;
    /* the operator () says to find the next command line option */
    while (( i = parser()) != EOF)
    {
	switch(i) 
	{
    	case 'D':
    	    Str = parser.optarg;	// read option
	    if (Str=="skip_dsjar")
	    {
		cout << "*** Debug: Skipping ds.jar\n";
		Config.Debug.skip_dsjar=true;
		break;
	    }
	    if (Str=="skip_lame")
	    {	
		cout << "*** Debug: Skipping lame\n";
		Config.Debug.skip_lame=true;
		break;
	    }
	    if (Str=="skip_mencoder")
	    {
		cout << "*** Debug: Skipping mencoder\n";
		Config.Debug.skip_mencoder=true;
		break;
	    }
	    if (Str=="nodelete")
	    {
		cout << "*** Debug: Not deleting any created files !\n";
		Config.Debug.nodelete=true;
		break;
	    }
	    if (Str=="help")
	    {
		UsageDebug();	// Leave Main Thread !!
		return 0;
	    }
	    cout << "*** Warning: Ignoring unknown debug option \"" << Str << "\" - see -D help for valid options!\n";
	    break;
        case 'd':
    	    cout << "*** Debug Level increased... not functional yet:( " << parser.optarg << "\n";
	    break;
	case '?':
	    Usage();
	    return 1;
        default:
            break;
	}
    }
// All options parsed, remaining arguments were sorted to the end of argv,
// starting with index parser.optind - if no non-options given,
// parser.optind points to argc !

    if (parser.optind>argc-1)	// more than one path specified
    {
	cerr << "*** Fatal: No Source Dir specified...\n";
	Usage();
	return 1;
    }
    if (parser.optind<argc-1)	// more than one path specified
    {
	cerr << "*** Fatal: Please only use one Source Dir at a time :-(\n";
	Usage();
	return 1;
    }
    else
    {
        SrcVideo.Path=argv[parser.optind];
// Make sure SrcVideo.Path ends upon ".../"
	if (SrcVideo.Path[strlen(SrcVideo.Path)-1]!='/')
	    SrcVideo.Path=SrcVideo.Path+"/";
	
// Check for existance of SrcVideo.Path
	if (!FileExist((const char*)SrcVideo.Path))
        {
	    cerr << "*** Fatal: Source Dir \"" << SrcVideo.Path << "\" could not be found ...\n";
	    return 0;
	}
// Check for existance of 001.vdr file
	FileName=SrcVideo.Path+"001.vdr";
	if (!FileExist((const char*)FileName))
	{
	    cerr << "*** Fatal: No 001.vdr File found - exiting\n";
	    return 0;
	}
	else
	    SrcVideo.FileName="001.vdr";
	    
// Determine Number of 00*.vdr File (up to 999.vdr - should be enough ;-)
	i=1;	
	while (i<=999)
	{
	    FileName=SrcVideo.Path+IntToString(i,3)+".vdr";
	    if (!FileExist((const char*) FileName))
	    {
		i--;
		break;
	    }
	    i++;
	} 
	SrcVideo.NumFiles=i;
	
	if (i>1)
	{
	    cout << "*** Info: Multiple Input Files (001.vdr up to " << IntToString(i,3) << ".vdr) found ...\n";
	}
		
// Check for existance of marks.vdr file
	FileName=SrcVideo.Path+"marks.vdr";
	if (!FileExist((const char*)FileName))
	{
	    cout << "*** Info: No marks.vdr File found - assuming default bitrate\n";
	}
// Read in marks.vdr File and analyze video length
	InFile.open (FileName, ifstream::in);
	while (InFile.good())
	{
	  Number="";
	  OneChar = (char) InFile.get();
	  while (OneChar!='\n' && InFile.good())
	  {
	    Number+=OneChar;
	    OneChar = (char) InFile.get();
	  }
	  // Here we have read in ONE line from marks.vdr and want to know
	  // the frames value !
	  if (strlen(Number)>=10 && strlen(Number)<=11)
	  {
	    // okay seems a valid line !
	//    cout << Number << "\n";
	    Str=Number.before(":");
	    Hours=atoi((const char *) Str);
	    Str2=Number.after(":");
	    Str=Str2.before(":");
	    Minutes=atoi((const char *) Str);
	    Str3=Str2.after(":");
	    Str=Str3.before(".");
	    Seconds=atoi((const char *) Str);
	    Str=Str3.after(".");
	    SingleFrames=atoi((const char *) Str);
	    SrcVideo.Marks[SrcVideo.NumMarks]=SingleFrames+25*Seconds+25*60*Minutes+25*60*60*Hours;
	    SrcVideo.NumMarks++;
	    if (SrcVideo.NumMarks==MaxMarks)
		break;
	  }

	  
	}
	if (SrcVideo.NumMarks>0)
	    SrcVideo.Frames=SrcVideo.Marks[SrcVideo.NumMarks-1];	//FIXME: Should consider in-/out marks ?
	InFile.close();
	
	DstVideo.Frames=SrcVideo.Frames; // we encode everything :-)
	
// Choose Media from BMM
	if (!SrcVideo.HasAC3)
	{
	// mp3
	    DstVideo.ABitrate=152;
	    if (DstVideo.Frames < (10*60*25))
	    {
		DstVideo.VBitrate = Config.DefaultBitrate;
		cout << "*** Since Video is shorter than 10 minutes, we assume Music Video Clip,\n";
		cout << "***   and will use a Videobitrate of " << DstVideo.VBitrate << "kbit/s.\n";
		DstVideo.NumFiles=1;
	    }
	    else		
	    {
		i=-1;
		do
	        {
	        i++;
	        VBitrate = CalcVBitrate (Config.BMM.CDSize[i], Config.BMM.NumCDs[i],
			                 Config.BMM.Files[i], SrcVideo.Frames,
				         DstVideo.ABitrate);
    		} while (VBitrate < Config.BMM.MVR_MP3[i] &&
		    		i < Config.BMM.Columns_MP3-1);
		if (VBitrate < Config.BMM.MVR_MP3[i] &&
			    i == Config.BMM.Columns_MP3-1)
		{
		    cout << "*** Warning: VBitrate drops below specified minimum value in MVR_MP3 !\n";
		    cout << "***   Encoding anyway using lower bitrate of " << VBitrate << "kbit/s\n";
		    cout << "***   Add further Media Combinations to 2divx.conf to avoid this ...\n";
		}
		DstVideo.VBitrate = VBitrate;
		DstVideo.Size = Config.BMM.CDSize[i] * Config.BMM.NumCDs[i] 
					/ Config.BMM.Files[i];
		DstVideo.NumFiles = Config.BMM.NumCDs[i];
	    
		cout << "*** Encoding with a Videobitrate of " << DstVideo.VBitrate <<
			"kbits/s - creating a file of ~" << DstVideo.Size << "MB\n";
		if (Config.BMM.Files[i] > 1)
	    	    cout << "*** With this trying to fit " << Config.BMM.Files[i] << 
		    	    " files on Media of " << Config.BMM.CDSize[i] <<
			    "MB Size !\n";
		if (DstVideo.NumFiles > 1)
		    cout << "***   afterwards splitting this file into " << DstVideo.NumFiles << " parts.\n";
	    }
	}
	else
	{
	    cerr << "*** Sorry: AC3 support not yet implemented :-( ...\n";
	    return 1;
	}
//	Debug: SrcVideo.Dump();

// Determine destination Name and Directory (depending on series name)

    // are we converting a file in VDRroot structure ?
	if (SrcVideo.Path.contains(Config.VDRroot,0)==0)
	{
	    cout << "*** Warning: Source Dir ain't a part of VDR root - no name matching !\n";
	    cout << "***   We use output_(...) as the destination name.\n";
	    DstVideo.Path=Config.DstDir;
	    Name="output";
	    Ext=".avi";
	}
	else	// yes we can try to guess the name !
	{
	    if (SrcVideo.Path[strlen(SrcVideo.Path)-1]=='/') // should always be true
		Str=SrcVideo.Path.at(0,strlen(SrcVideo.Path)-1);
	    else
		Str=SrcVideo.Path;
	    Str=Str.after(Config.VDRroot); // remove VDR root
	    // we reverse the stuff in Str to work from end to beginning !
	    Str2=reverse(Str);
	    Str2=Str2.after('/'); // strip off yyyy-mm-dd.hh:mm.xx.xx.rec
	    Str3=reverse(Str2.before('/')); // extract FileName (Episode)
	    DstVideo.Path=Config.DstDir+reverse(Str2.after('/'))+'/';
	    if (Str3[0]=='%')	//strip % sign from cutting!
		Str3=Str3.after('%');
		
		// here we should do episode stuff !!!! with Str3 !!!
		
	    Name=Str3+"_("+IntToString(DstVideo.VBitrate,0)+"_lavc)";
	    Ext=".avi";
	}

// Make sure we find a unique filename for destination !

	if (FileExist(DstVideo.Path+Name+Ext))
	{
	    i=1;
	    Str=Name;
	    while (FileExist(DstVideo.Path+Str+Ext))
	    {
	        Str=Name+"_"+IntToString(i,0);
	        i++;		
	    }
	    Name=Str;
	}

    // now Name+Dst should contain real unique filename !

	DstVideo.FileName=Name+Ext;
	

// Start Demuxer	
	if (!Config.Debug.skip_dsjar)
	{
	    if (!dmx_dsjar(Config, SrcVideo))
	    {
		cerr << "*** Fatal: Major error during ds.jar - see above for details!\n";
		return 1;	
	    }
	}
//	SrcVideo.Dump();

// Starting Lame to transcode 001.mp2 to mp3 !
	if (!Config.Debug.skip_lame)
	{
	    if (!aen_lame(Config, SrcVideo))
	    {
		cerr << "*** Fatal: Major error during lame - see above for details!\n";
		return 1;	
	    }
	}

// Fire up Mencoder to reencode Video and read in freshly encoded audio stream	
	if (!Config.Debug.skip_mencoder)
	{
	    if (!ven_mencoder(Config, SrcVideo, DstVideo))
	    {
		cerr << "*** Fatal: Major error during mencoder - see above for details!\n";
		return 1;	
	    }
	}

// Check for destination Dir / create if needed
// Move AVI File to Destination

	if (!FileExist(DstVideo.Path))
	{
	    cout << "*** Info: Destination Path \"" << DstVideo.Path << "\" not found - creating\n";
	    i=system("mkdir -p \""+DstVideo.Path+"\"");
	    if (i!=0)
	    {
		cerr << "*** Warning: Destination Dir could not be created - trying default!\n";
		if (!FileExist(Config.DstDir))
		{
		    cerr << "*** Default Dir specified in 2divx.conf \"" << Config.DstDir << "\" isn't available :(\n";
		    return 1;		
		}
		else
		    DstVideo.Path=Config.DstDir; // we put all in DstDir root!
	    }
	}
	// here we made sure that DstVideo.Path does exist !
// finally move output avi to DstVideo.Path

	i=system("mv \""+Config.TempDir+"test.avi\" "+
		    "\""+DstVideo.Path+DstVideo.FileName+"\"");
	if (i!=0)
	{
		cerr << "*** Fatal: Error " << i << " occured during move of Destination file\n";
		cerr << "***   Check for sufficient rights in TempDir and DstDir !!!\n";
		return 1;
	}		

		
//	SrcVideo.Dump();	
//	DstVideo.Dump();


    }	
	// End of main thread
    cout << "*** done.\n";
    return 0;
}


// needed functions

bool CheckEnv ()
// Checks the environment of 2divx (Config files / external Programs / stuff)
// return true, if env seems to be acceptable :-)
{
    if (!FileExist("2divx.conf"))
    {
	cerr << "*** Fatal: 2divx.conf File not found in current path ...\n";
	return false;
    }
    

return true;
}

int CalcVBitrate(int CDSize, int NumCDs, int Files, int frames, int ABitrate)
{
    int VBitrate;
    VBitrate = (CDSize * NumCDs / Files * 1024 * 8 / (frames / 25)) - ABitrate;
    if (VBitrate > 6000) VBitrate=6000;
    return VBitrate;
}

void Usage(void)
{
    cout << "*** Usage: 2divx [-d -D debugfunction] /path/to/vdr/directory[/]\n";
    cout << "***   -d	Increase debug message level\n";
    cout << "*** 	(use -d -d for verbose debug mode!)\n";
    cout << "***   -D debugfunction	Special Debug Modes ! See -D help\n";
    cout << "\n";
    return;
}

void UsageDebug(void)
{
    cout << "*** Usage for -D Switch:\n";
    cout << "***  2divx -D [ skip_dsjar | skip_lame | skip_mencoder | nodelete ]\n";
    cout << "*** The skip_* Skips the corresponding module, however make sure the files\n";
    cout << "***  needed from the former process must still be present !\n";
    cout << "*** nodelete - simply doesn't delete any files 2divx processes\n";
    cout << "\n";
    return;
}    





